#include <vector>
#include <array>
#include <cassert>
#include <iostream>
#include <algorithm>

#include "production.h"
#include "Matrix.h"

using namespace std;
using namespace Grammar;

// shortcut for std::vector::contains()
template<class Container>
bool contains(const Container &c, const typename Container::value_type & v) {
    if(std::find(std::begin(c), std::end(c), v) != std::end(c)) {
        return true;
    }
    return false;
}

int main() {
    // not terminal
    NotTerminal N[] = {'S', 'A', 'B', 'C'};
    // terminal
    Terminal t[] = {'a', 'b'};
    // start symbol
    const Symbol startSynbol = 'S';
    // productions
    vector<Production> P;
    P.push_back(Production('S', {'A', 'B'}));
    P.push_back(Production('S', {'B', 'C'}));
    P.push_back(Production('A', {'B', 'A'}));
    P.push_back(Production('A', {'a'}));
    P.push_back(Production('B', {'C', 'C'}));
    P.push_back(Production('B', {'b'}));
    P.push_back(Production('C', {'A', 'B'}));
    P.push_back(Production('C', {'a'}));

    // word to check
    vector<Terminal> word = {'b', 'b', 'a', 'b', 'a', 'a'};

    const size_t n = word.size();
    Matrix<vector<Symbol>> V(word.size(), word.size());

    cout << "Productions:\n";
    for(const Production &p : P) {
        cout << p.left << " -> {";
        for(const auto &s : p.right) {
            cout << s << " ";
        }
        cout << "}\n";
    }

    cout << "Word to detect\n";
    for(Terminal t : word) {
        cout << t << "\t";
    }
    cout << "\n";

    /*
    Für i = 1 ... n
      Für jede Produktion (l \rightarrow r) \in P
        Falls r = w_i
          Setze V_{i,1}:=V_{i,1} \cup \{ l \}
    */

    for(size_t col=0; col < n; col++) {
        const Terminal &wi = word.at(col);
        for(const Production &p : P) {
            if(contains(p.right, wi)) {
                V.at(0, col).push_back(p.left);
            }
        }
    }


    /*
    Für j = 2 ... n
      Für i = 1 ... n - j + 1
        Für k = 1 ... j - 1
          Setze V_{i,j} := V_{i,j} + { l | \exists B,C in N ((l -> BC) in P and B in V_{i,k} and C in V_{i + k, j - k}) \}

    Falls S \in V_{1,n}, stoppe und gib "w wird von G erzeugt" aus
    Stoppe und gib "w wird nicht von G erzeugt" aus
    */
    for(size_t row=1; row < n; row++) {
//        cout << "row " << row << "\n";
        for(size_t col=0; col < n-row; col++) {
//            cout << "col " << col << "\n";
            for(size_t k=0; k < row; k++) {
                const auto &e1 = V.at(k, col);
                const auto &e2 = V.at(row-k-1, col+k+1);

                for(size_t i=0; i < e1.size(); i++) {
                    for(size_t j=0; j < e2.size(); j++) {
//                        cout << "Check " << e1.at(i) << " " << e2.at(j) << "\n";
                        const Production::String toTest({e1.at(i), e2.at(j)});
                        for(const Production& p : P) {
                            if(p.right == toTest) {
                                // uniq append
                                if(contains(V.at(row, col), p.left) == false) {
                                    V.at(row, col).push_back(p.left);
                                }
                            }
                        }
                    }
                }
            }
        }
    }

    cout << "Matrix\n";
    V.print();
    cout << "\n";

    if(contains(V.at(n-1, 0), startSynbol)) {
        cout << "word recognized\n";
    }
}
